# _  __  ____    _   _
# | |/ / |  _ \  | \ | |
# | ' /  | |_) | |  \| |
# | . \  |  __/  | |\  |
# |_|\_\ |_|     |_| \_|
#
# (c) 2018 KPN
# License: MIT license.
# Author: Jan Bogaerts
#
# sensor names

from kpn_senml.senml_unit import enum


SenmlNames = enum(KPN_SENML_PRESSURE="pressure",
                KPN_SENML_ANGLE="angle",
                KPN_SENML_LENGHT="lenght",
                KPN_SENML_BREADTH="breadth",
                KPN_SENML_HEIGHT="height",
                KPN_SENML_WEIGHT="weight",
                KPN_SENML_THICKNESS="thickness",
                KPN_SENML_DISTANCE="distance",
                KPN_SENML_AREA="area",
                KPN_SENML_VOLUME="volume",
                KPN_SENML_VELOCITY="velocity",
                KPN_SENML_ELECTRICCURRENT="electricCurrent",
                KPN_SENML_ELECTRICPOTENTIAL="electricPotential",
                KPN_SENML_ELECTRICRESISTANCE="electricResistance",
                KPN_SENML_TEMPERATURE="temperature",
                KPN_SENML_ILLUMINANCE="illuminance",
                KPN_SENML_ALTITUDE="altitude",
                KPN_SENML_ACCELERATIONX="accelerationX",
                KPN_SENML_ACCELERATIONY="accelerationY",
                KPN_SENML_ACCELERATIONZ="accelerationZ",
                KPN_SENML_HEADING="heading",
                KPN_SENML_LONGITUDE="longitude",
                KPN_SENML_LATTITUDE="lattitude",
                KPN_SENML_CARBONMONOXIDE="carbonMonoxide",
                KPN_SENML_CARBONDIOXIDE="carbonDioxide",
                KPN_SENML_SOUND="sound",
                KPN_SENML_FREQUENCY="frequency",
                KPN_SENML_BATTERYLEVEL="batteryLevel",
                KPN_SENML_HUMIDITY="humidity")


# _  __  ____    _   _
# | |/ / |  _ \  | \ | |
# | ' /  | |_) | |  \| |
# | . \  |  __/  | |\  |
# |_|\_\ |_|     |_| \_|
#
# (c) 2018 KPN
# License: MIT license.
# Author: Jan Bogaerts
#
# base for senml objects

class SenmlBase(object):
    '''
    the base class for all senml objects.
    '''


# _  __  ____    _   _
# | |/ / |  _ \  | \ | |
# | ' /  | |_) | |  \| |
# | . \  |  __/  | |\  |
# |_|\_\ |_|     |_| \_|
#
# (c) 2018 KPN
# License: MIT license.
# Author: Jan Bogaerts
#
# pack object


from kpn_senml.senml_record import SenmlRecord
from kpn_senml.senml_base import SenmlBase
import json
import cbor2


class SenmlPackIterator:
    '''an iterator to walk over all records in a pack'''

    def __init__(self, list):
        self._list = list
        self._index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if(self._index > len(self._list)):
            res = self._list[self._index]
            self._index += 1
            return res
        else:
            raise StopIteration()

class SenmlPack(SenmlBase):
    '''
    represents a senml pack object. This can contain multiple records but also other (child) pack objects.
    When the pack object only contains records, it represents the data of a device.
    If the pack object has child pack objects, then it represents a gateway
    '''

    json_mappings = {'bn': 'bn', 'bt': 'bt', 'bu': 'bu', 'bv': 'bv', 'bs': 'bs',
                      'n': 'n', 'u': 'u', 'v': 'v', 'vs': 'vs', 'vb': 'vb', 'vd': 'vd', 's': 's', 't': 't', 'ut': 'ut'}

    def __init__(self, name, callback=None):
        '''
        initialize the object
        :param name: {string} the name of the pack
        '''
        self._data = []
        self.name = name
        self._base_value = None
        self._base_time = None
        self._base_sum = None
        self.base_unit = None
        self._parent = None                                 # a pack can also be the child of another pack.
        self.actuate = callback                             # actuate callback function

    def __iter__(self):
        return SenmlPackIterator(self._data)

    def __enter__(self):
        '''
        for supporting the 'with' statement
        :return: self
        '''
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        '''
        when destroyed in a 'with' statement, make certain that the item is removed from the parent list.
        :return: None
        '''
        if self._parent:
            self._parent.remove(self)

    @property
    def base_value(self):
        '''
        the base value of the pack. The value of the records will be subtracted by this value during rendering.
        While parsing, this value is added to the value of the records.
        :return: a number
        '''
        return self._base_value

    @base_value.setter
    def base_value(self, value):
        '''
        set the base value.
        :param value: only number allowed
        :return:
        '''
        self._check_value_type(value, "base_value")
        self._base_value = value


    @property
    def base_sum(self):
        '''
        the base sum of the pack.
        :return: a number
        '''
        return self._base_sum

    @base_sum.setter
    def base_sum(self, value):
        '''
        set the base sum.
        :param value: only number allowed
        :return:
        '''
        self._check_value_type(value, "base_sum")
        self._base_sum = value

    @property
    def base_time(self):
        '''
        Get the base time assigned to this pack object.
        While rendering, this value will be subtracted from the value of the records.
        :return: unix time stamp representing the base time
        '''
        return self._base_time

    @base_time.setter
    def base_time(self, value):
        self._check_value_type(value, "base_time")
        self._base_time = value

    def _check_value_type(self, value, field_name):
        '''
        checks if the type of value is allowed for senml
        :return: None, raisee exception if not ok.
        '''
        if (not value == None):
            if not (isinstance(value, int) or isinstance(value, float)):
                raise Exception("invalid type for " + field_name + ", only numbers allowed")


    def from_json(self, data):
        '''
        parse a json string and convert it to a senml pack structure
        :param data: a string containing json data.
        :return: None, will call the appropriate callback functions.
        '''
        records = json.loads(data)                                              # load the raw senml data
        self._process_incomming_data(records, SenmlPack.json_mappings)


    def _process_incomming_data(self, records, naming_map):
        '''
        generic processor for incoming data (actuators.
        :param records: the list of raw senml data, parsed from a json or cbor structure
        :param naming_map: translates cbor to json field names (when needed).
        :return: None
        '''
        cur_pack_el = self
        new_pack = False
        for item in records:
            if naming_map['bn'] in item:                                                    # ref to a pack element, either this or a child pack.
                if item[naming_map['bn']] != self.name:
                    pack_el = next( (x for x in self._data if x.name == item[naming_map['bn']]), None )
                else:
                    pack_el = self
                if pack_el:
                    cur_pack_el = pack_el
                    new_pack = False
                else:
                    device = SenmlPack(item[naming_map['bn']])
                    self._data.append(device)
                    cur_pack_el = device
                    new_pack = True

                if naming_map['bv'] in item:                                        # need to copy the base value assigned to the pack element so we can do proper conversion for actuators.
                    cur_pack_el.base_value = item[naming_map['bv']]

                rec_el = next( (x for x in cur_pack_el._data if x.name == item[naming_map['n']]), None )
                if rec_el:
                    rec_el.do_actuate(item, naming_map)
                elif new_pack:
                    self.do_actuate(item, naming_map, cur_pack_el)
                else:
                    cur_pack_el.do_actuate(item, naming_map)
            else:
                rec_el = next((x for x in self._data if x.name == item[naming_map['n']]), None)
                if rec_el:
                    rec_el.do_actuate(item, naming_map)
                elif new_pack:
                    self.do_actuate(item, naming_map, cur_pack_el)
                else:
                    cur_pack_el.do_actuate(item, naming_map)


    def do_actuate(self, raw, naming_map, device = None):
        '''
        called while parsing incoming data for a record that is not yet part of this pack object.
        adds a new record and raises the actuate callback of the pack with the newly created record as argument
        :param naming_map:
        :param device: optional: if the device was not found
        :param raw: the raw record definition, as found in the json structure. this still has invalid labels.
        :return: None
        '''
        rec = SenmlRecord(raw[naming_map['n']])
        if device:
            device.add(rec)
            rec._from_raw(raw, naming_map)
            if self.actuate:
                self.actuate(rec, device=device)
        else:
            self.add(rec)
            rec._from_raw(raw, naming_map)
            if self.actuate:
                self.actuate(rec, device=None)


    def to_json(self):
        '''
        render the content of this object to a string.
        :return: a string representing the senml pack object
        '''
        converted = []
        self._build_rec_dict(SenmlPack.json_mappings, converted)
        return json.dumps(converted)


    def _build_rec_dict(self, naming_map, appendTo):
        '''
        converts the object to a senml object with the proper naming in place.
        This can be recursive: a pack can contain other packs.
        :param naming_map: a dictionary used to pick the correct field names for either senml json or senml cbor
        :return: None
        '''
        internalList = []
        for item in self._data:
            item._build_rec_dict(naming_map,internalList)
        if len(internalList) > 0:
            first_rec = internalList[0]
        else:
            first_rec = {}
            internalList.append(first_rec)

        if self.name:
            first_rec[naming_map['bn']] = self.name
        if self.base_value:
            first_rec[naming_map['bv']] = self.base_value
        if self.base_unit:
            first_rec[naming_map['bu']] = self.base_unit
        if self.base_sum:
            first_rec[naming_map['bs']] = self.base_sum
        if self.base_time:
            first_rec[naming_map['bt']] = self.base_time
        appendTo.extend(internalList)

    def from_cbor(self, data):
        '''
        parse a cbor data byte array to a senml pack structure.
        :param data: a byte array.
        :return: None
        '''
        records = cbor2.loads(data)  # load the raw senml data
        naming_map = {'bn': -2, 'bt': -3, 'bu': -4, 'bv': -5, 'bs': -16,
                      'n': 0, 'u': 1, 'v': 2, 'vs': 3, 'vb': 4, 'vd': 8, 's': 5, 't': 6, 'ut': 7}
        self._process_incomming_data(records, naming_map)

    def to_cbor(self):
        '''
        render the content of this object to a cbor byte array
        :return: a byte array
        '''
        naming_map = {'bn': -2, 'bt': -3, 'bu': -4, 'bv': -5, 'bs': -16,
                      'n': 0, 'u': 1, 'v': 2, 'vs': 3, 'vb': 4, 'vd': 8, 's': 5, 't': 6, 'ut': 7}
        converted = []
        self._build_rec_dict(naming_map, converted)
        return cbor2.dumps(converted)

    def add(self, item):
        '''
        adds the item to the list of records
        :param item: {SenmlRecord} the item that needs to be added to the pack
        :return: None
        '''
        if not (isinstance(item, SenmlBase)):
            raise Exception('invalid type of param, SenmlRecord or SenmlPack expected')
        if not item._parent == None:
            raise Exception('item is already part of a pack')

        self._data.append(item)
        item._parent = self

    def remove(self, item):
        '''
        removes the item from the pack
        :param item: {SenmlRecord} the item that needs to be removed
        :return: None
        '''
        if not (isinstance(item, SenmlBase)):
            raise Exception('invalid type of param, SenmlRecord or SenmlPack expected')
        if not item._parent == self:
            raise Exception('item is not part of this pack')

        self._data.remove(item)
        item._parent = None

    def clear(self):
        '''
        clear the list of the pack
        :return: None
        '''
        for item in self._data:
            item._parent = None
        self._data = []


# _  __  ____    _   _
# | |/ / |  _ \  | \ | |
# | ' /  | |_) | |  \| |
# | . \  |  __/  | |\  |
# |_|\_\ |_|     |_| \_|
#
# (c) 2018 KPN
# License: MIT license.
# Author: Jan Bogaerts
#
# unit  names

def enum(**enums):
    return type('Enum', (), enums)

SenmlUnits = enum(SENML_UNIT_METER="m",
                SENML_UNIT_KILOGRAM="kg",
                SENML_UNIT_GRAM="g",
                SENML_UNIT_SECOND="s",
                SENML_UNIT_AMPERE="A",
                SENML_UNIT_KELVIN="K",
                SENML_UNIT_CANDELA="cd",
                SENML_UNIT_MOLE="mol",
                SENML_UNIT_HERTZ="Hz",
                SENML_UNIT_RADIAN="rad",
                SENML_UNIT_STERADIAN="sr",
                SENML_UNIT_NEWTON="N",
                SENML_UNIT_PASCAL="Pa",
                SENML_UNIT_JOULE="J",
                SENML_UNIT_WATT="W",
                SENML_UNIT_COULOMB="C",
                SENML_UNIT_VOLT="V",
                SENML_UNIT_FARAD="F",
                SENML_UNIT_OHM="Ohm",
                SENML_UNIT_SIEMENS="S",
                SENML_UNIT_WEBER="Wb",
                SENML_UNIT_TESLA="T",
                SENML_UNIT_HENRY="H",
                SENML_UNIT_DEGREES_CELSIUS="Cel",
                SENML_UNIT_LUMEN="lm",
                SENML_UNIT_LUX="lx",
                SENML_UNIT_BECQUEREL="Bq",
                SENML_UNIT_GRAY="Gy",
                SENML_UNIT_SIEVERT="Sv",
                SENML_UNIT_KATAL="kat",
                SENML_UNIT_SQUARE_METER="m2",
                SENML_UNIT_CUBIC_METER="m3",
                SENML_UNIT_LITER="l",
                SENML_UNIT_VELOCITY="m/s",
                SENML_UNIT_ACCELERATION="m/s2",
                SENML_UNIT_CUBIC_METER_PER_SECOND="m3/s",
                SENML_UNIT_LITER_PER_SECOND="l/s",
                SENML_UNIT_WATT_PER_SQUARE_METER="W/m2",
                SENML_UNIT_CANDELA_PER_SQUARE_METER="cd/m2",
                SENML_UNIT_BIT="bit",
                SENML_UNIT_BIT_PER_SECOND="bit/s",
                SENML_UNIT_DEGREES_LATITUDE="lat",
                SENML_UNIT_DEGREES_LONGITUDE="lon",
                SENML_UNIT_PH="pH",
                SENML_UNIT_DECIBEL="db",
                SENML_UNIT_DECIBEL_RELATIVE_TO_1_W="dBW",
                SENML_UNIT_BEL="Bspl",
                SENML_UNIT_COUNTER="count",
                SENML_UNIT_RATIO="//",
                SENML_UNIT_RELATIVE_HUMIDITY="%RH",
                SENML_UNIT_PERCENTAGE_REMAINING_BATTERY_LEVEL="%EL",
                SENML_UNIT_SECONDS_REMAINING_BATTERY_LEVEL="EL",
                SENML_UNIT_EVENT_RATE_PER_SECOND="1/s",
                SENML_UNIT_EVENT_RATE_PER_MINUTE="1/min",
                SENML_UNIT_BPM="beat/min",
                SENML_UNIT_BEATS="beats",
                SENML_UNIT_SIEMENS_PER_METER="S/m")


# _  __  ____    _   _
# | |/ / |  _ \  | \ | |
# | ' /  | |_) | |  \| |
# | . \  |  __/  | |\  |
# |_|\_\ |_|     |_| \_|
#
# (c) 2018 KPN
# License: MIT license.
# Author: Jan Bogaerts
#
# record object

import base64
from kpn_senml.senml_base import SenmlBase


class SenmlRecord(SenmlBase):
    '''represents a single value in a senml pack object'''

    def __init__(self, name, **kwargs):
        '''
        create a new senml record
        :param kwargs:  optional parameters:
                            - value: the value to store in the record
                            - time: the timestamp to use (when was the value measured)
                            - name: the name of hte record
                            - unit: unit value
                            - sum: sum value
                            - update_time: max time before sensor will provide an updated reading
                            - callback: a callback function taht will be called when actuator data has been found. Expects no params
        '''
        self.__parent = None                                    # using double __ cause it's a field for an internal property
        self._unit = None                                       # declare and init internal fields
        self._value = None
        self._time = None
        self._sum = None
        self._update_time = None

        self._parent = None                                     # internal reference to the parent object
        self.name = name
        self.unit = kwargs.get('unit', None)
        self.value = kwargs.get('value',None)
        self.time = kwargs.get('time',None)
        self.sum = kwargs.get('sum', None)
        self.update_time = kwargs.get('update_time', None)
        self.actuate = kwargs.get('callback', None)          # actuate callback function


    def __enter__(self):
        '''
        for supporting the 'with' statement
        :return: self
        '''
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        '''
        when destroyed in a 'with' statement, make certain that the item is removed from the parent list.
        :return: None
        '''
        if self._parent:
            self._parent.remove(self)

    def _check_value_type(self, value):
        '''
        checks if the type of value is allowed for senml
        :return: None, raisee exception if not ok.
        '''
        if (not value == None):
            if not (isinstance(value, bool) or isinstance(value, int) or
                    isinstance(value, float) or isinstance(value, bytearray) or
                    isinstance(value, basestring)):
                raise Exception("invalid type for value, only numbers, strings, boolean and byte arrays allowed")

    def _check_number_type(self, value, field_name):
        '''
        checks if the type of value is allowed for senml
        :return: None, raisee exception if not ok.
        '''
        if (not value == None):
            if not (isinstance(value, int) or isinstance(value, float)):
                raise Exception("invalid type for " + field_name + ", only numbers allowed")


    @property
    def value(self):
        '''get the value currently assigned to the object'''
        return self._value

    @value.setter
    def value(self, value):
        '''set the current value. Will not automatically update the time stamp. This has to be done seperatly for more
            finegrained control
            Note: when the value is a float, you can control rounding in the rendered output by using the function
            round() while assigning the value. ex: record.value = round(12.2 / 1.5423, 2)
        '''
        self._check_value_type(value)
        self._value = value


    @property
    def time(self):
        ''' get the time at which the measurement for the record was taken.
        :return: a unix time stamp. This is the absolute value, not adjusted to the base time of the pack.
        '''
        return self._time

    @time.setter
    def time(self, value):
        '''
        Set the time at which the measurement for this record was taken.
        This should be the absolute value. Any adjustments that need to be done for the base-time of the pack are
        done automatically during rendering.
        :param value: a unix time stamp
        :return: None.
        '''
        self._check_number_type(value, 'time')
        self._time = value

    @property
    def update_time(self):
        ''' get the time at which the next measurement is expected to be taken for this record.
        :return: a unix time stamp. This is the absolute value, not adjusted to the base time of the pack.
        '''
        return self._update_time

    @update_time.setter
    def update_time(self, value):
        '''
        Set the time at which the next measurement is expected to be taken for this record.
        This should be the absolute value. Any adjustments that need to be done for the base-time of the pack are
        done automatically during rendering.
        :param value: a unix time stamp
        :return: None.
        '''
        self._check_number_type(value, 'update_time')
        self._update_time = value

    @property
    def sum(self):
        return self._sum

    @sum.setter
    def sum(self, value):
        self._check_number_type(value, 'sum')
        self._sum = value

    @property
    def _parent(self):
        '''
        the parent pack object for this record. This is a property so that inheriters can override and do custom
        actions when the parent is set (like passing it on to their children
        :return: a SenMLBase object that contains this object as child.
        '''
        return self.__parent

    @_parent.setter
    def _parent(self, value):
        '''
        Set the parent pack object for this record. This is a property so that inheriters can override and do custom
        actions when the parent is set (like passing it on to their children
        :param value: the object to assign as parent.
        :return: None
        '''
        self.__parent = value

    def _build_rec_dict(self, naming_map, appendTo):
        '''
        converts the object to a dictionary that can be rendered to senml.
        :param naming_map: a dictionary that maps the field names to senml json or senml cbor. keys are in the
        form 'n', 'v',...  values for 'n' are either 'n' or 0 (number is for cbor)
        :return: a senml dictionary representation of the record
        '''
        result = { }

        if self.name:
            result[naming_map['n']] = self.name

        if self._sum:
            if self._parent and self._parent.base_sum:
                result[naming_map['s']] = self._sum - self._parent.base_sum
            else:
                result[naming_map['s']] = self._sum
        elif isinstance(self._value, bool):
            result[naming_map['vb']] = self._value
        elif isinstance(self._value, int) or isinstance(self._value, float):
            if self._parent and self._parent.base_value:
                result[naming_map['v']] = self._value - self._parent.base_value
            else:
                result[naming_map['v']] = self._value
        elif isinstance(self._value, basestring):
            result[naming_map['vs']] = self._value
        elif isinstance(self._value, bytearray):
            if naming_map['vd'] == 'vd':                # neeed to make a distinction between json (needs base64) and cbor (needs binary)
                result[naming_map['vd']] = base64.b64encode(self._value)
            else:
                result[naming_map['vd']] = self._value
        else:
            raise Exception("sum or value of type bootl, number, string or byte-array is required")

        if self._time:
            if self._parent and self._parent.base_time:
                result[naming_map['t']] = self._time - self._parent.base_time
            else:
                result[naming_map['t']] = self._time

        if self.unit:
            result[naming_map['u']] = self.unit

        if self._update_time:
            if self._parent and self._parent.base_time:
                result[naming_map['ut']] = self._update_time - self._parent.base_time
            else:
                result[naming_map['ut']] = self._update_time

        appendTo.append(result)


    def _from_raw(self, raw, naming_map):
        '''
        extracts te data from the raw record. Used during parsing of incoming data.
        :param raw: a raw senml record which still contains the original field names
        :param naming_map: used to map cbor names to json field names
        :return: None
        '''
        if naming_map['v'] in raw:
            val = raw[naming_map['v']]
            if self._parent and self._parent.base_value:
                val += self._parent.base_value
        elif naming_map['vs'] in raw:
            val = raw[naming_map['vs']]
        elif naming_map['vb'] in raw:
            val = raw[naming_map['vb']]
        elif naming_map['vd'] in raw:
            val = base64.b64decode(raw[naming_map['vb']])
        else:
            val = None
        self.value = val

    def do_actuate(self, raw, naming_map):
        '''
        called when a raw senml record was found for this object. Stores the data and if there is a callback, calls it.
        :param raw: raw senml object
        :return: None
        '''
        self._from_raw(raw, naming_map)
        if self.actuate:
            self.actuate(self)


# _  __  ____    _   _
# | |/ / |  _ \  | \ | |
# | ' /  | |_) | |  \| |
# | . \  |  __/  | |\  |
# |_|\_\ |_|     |_| \_|
#
# (c) 2018 KPN
# License: MIT license.
# Author: Jan Bogaerts
#
# include all

from kpn_senml.senml_base import SenmlBase
from .senml_pack import SenmlPack
from .senml_record import SenmlRecord
from .senml_unit import SenmlUnits
from .senml_kpn_names import SenmlNames

'''
The KPN SenML library helps you create and parse [senml documents](https://tools.ietf.org/html/draft-ietf-core-senml-13) 
in both json and cbor format.
'''


from setuptools import setup

setup(
    name='kpn_senml',
    version='1.0.0',
    packages=['kpn_senml'],
    url='https://github.com/jan-bogaerts/slate',
    license='APACHE',
    author='Jan Bogaerts',
    author_email='jb@elastetic.com',
    description='generate and parse senml json and cbor data',
    long_description='With this library you can generate senml packs containing sensor data in both json and cbor format.  It can also parse senml data in both json and cbor format in order to support actuators.',
    keywords='senml kpn cbor json',
    install_requires=[
          'cbor2',
    ]
)
